-----------------------------------------------------------------------------
-- eval.e
-- An expression evaluation library for Euphoria.  (compatible with v2.0)
-- by Falkon1313
-- 26 June, 1999 - 02 July, 1999
-----------------------------------------------------------------------------

include get.e  --value()


global constant EVAL_SUCCESS = 1, EVAL_ERROR = 0


constant WHITESPACE = "\n\t\r ",

	 UNARY_OPS = { "+", "-", "not", "!" },

	 OPERATORS = UNARY_OPS &
		     { "^",
		       "*", "/", "%",
		       "&",
		       "<", ">", "=", "<=", ">=", "!=", "==",
		       "&&", "||", "!|",
		       "and", "or"
		      },
	    PLUS = 1,
	    MINUS = 2,
	    NOT = 3,
	    BITWISE_NOT = 4,
	    POWER = 5,
	    TIMES = 6,
	    DIVIDED_BY = 7,
	    MOD = 8,
	    CONCATENATE = 9,
	    LESS_THAN = 10,
	    GREATER_THAN = 11,
	    EQUAL_TO = 12,
	    LESS_OR_EQUAL = 13,
	    GREATER_OR_EQUAL = 14,
	    NOT_EQUAL = 15,
	    COMPARISON = 16,
	    BITWISE_AND = 17,
	    BITWISE_OR = 18,
	    BITWISE_XOR = 19,
	    AND = 20,
	    OR = 21,

	 PRECEDENCE =
	    { { POWER },
	      { TIMES, DIVIDED_BY, MOD },
	      { PLUS, MINUS },
	      { CONCATENATE },
	      { LESS_THAN, GREATER_THAN, EQUAL_TO, LESS_OR_EQUAL,
		GREATER_OR_EQUAL, NOT_EQUAL, COMPARISON },
	      { BITWISE_AND, BITWISE_OR, BITWISE_XOR },
	      { AND, OR }
	    },

	 SAFEOPS = { CONCATENATE, COMPARISON }



-----------------------------------------------------------------------------
-- Minor functions
-----------------------------------------------------------------------------

function recfind( object val, object obj )
   --Boolean: recurses down through subsequences until it finds a value
   --returns 1 if found, 0 if not found

   if atom( obj ) then
      return( not compare( val, obj ) )
   end if
   for count = 1 to length( obj ) do
      if compare( val, obj[count] ) = 0 then
	 return( 1 )
      elsif sequence( obj[count] ) then
	 if recfind( val, obj[count] ) then
	    return( 1 )
	 end if
      end if
   end for
   return( 0 )
end function


function separator( integer ch )
   --tells whether a character is a 'separator',
   --or a legal part of an identifier or literal
   if find( ch, "#._0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz" ) then
      return( 0 )
   end if
   return( 1 )
end function


function numeric_literal( sequence s )
   --s is a possible numeric literal in string form
   --either a hexadecimal, decimal, or binary number
   --returns {EVAL_SUCCESS, value} or {EVAL_ERROR, 0}
   object val

   if s[1] = '#' then
      --hexadecimal?
      for c = 2 to length( s ) do
	 if not find( s[c], "1234567890ABCDEF" ) then
	    --Not a valid numeric literal
	    return( { EVAL_ERROR, 0 } )
	 end if
      end for
      val = value( s )  -- euphoria-style hexadecimal works with value
      val = val[2]
   elsif s[length( s )] = 'h' then
      --hexadecimal?
      for c = 1 to length( s ) - 1 do
	 if not find( s[c], "1234567890ABCDEF" ) then
	    --Not a valid numeric literal
	    return( { EVAL_ERROR, 0 } )
	 end if
      end for
      val = value( '#' & s[1..length( s ) - 1] )
      val = val[2]
   elsif s[length( s )] = 'b' then
      --binary?
      for c = 1 to length( s ) - 1 do
	 if not find( s[c], "10" ) then
	    --Not a valid numeric literal
	    return( { EVAL_ERROR, 0 } )
	 end if
      end for
      val = 0
      for c = 1 to length( s ) - 1 do
	 val = ( val * 2 ) + ( s[c] = '1' )
      end for
   else
      --decimal?
      for c = 1 to length( s ) do
	 if not find( s[c], "1234567890." ) then
	    --Not a valid numeric literal
	    return( { EVAL_ERROR, 0 } )
	 end if
      end for
      val = value( s )
      val = val[2]
   end if

   return( { EVAL_SUCCESS, val } )
end function


function string_or_character_literal( sequence s, integer q )
   --s is a possible string or character literal in string form
   --q is the quote type, '"' for string, '\'' for character
   --returns {success, value}
   object val
   if q = '\'' then
      --char literal must be 3 or 4 characters
      if length( s ) < 3 or length( s ) > 4 then
	 return( { EVAL_ERROR, 0 } )
      end if
   elsif length( s ) < 2 then
      --string literal must be 2 or more characters
      return( { EVAL_ERROR, 0 } )
   end if

   if s[1] = q and s[length( s )] = q then
      val = value( s )
      if val[1] = GET_SUCCESS then
	 return( { EVAL_SUCCESS, val[2] } )
      end if
   end if
   return( { EVAL_ERROR, 0 } )
end function


function operate( object o1, integer operator, object o2 )
   --handles a binary operation.
   --returns {EVAL_SUCCESS, result} or {EVAL_ERROR, errormsg}
   --o1 is first operand, o2 is second operand

   --check for sequence lengths to prevent crash in arithmetical operations
   if not find( operator, SAFEOPS ) then
      if sequence( o1 ) and sequence( o2 ) then
	 if length( o1 ) != length( o2 ) then
	    return( { EVAL_ERROR, sprintf( "Error: Sequence lengths not the same. (%d != %d)", { length( o1 ), length( o2 ) } ) } )
	 end if
      end if
   end if

   if operator = PLUS then
      return( { EVAL_SUCCESS, o1 + o2 } )
   elsif operator = MINUS then
      return( { EVAL_SUCCESS, o1 - o2 } )
   elsif operator = TIMES then
      return( { EVAL_SUCCESS, o1 * o2 } )
   elsif operator = DIVIDED_BY then
      if recfind( 0, o2 ) then
	 return( { EVAL_ERROR, "ERROR: Attempt to divide by zero." } )
      end if
      return( { EVAL_SUCCESS, o1 / o2 } )
   elsif operator = LESS_THAN then
      return( { EVAL_SUCCESS, o1 < o2 } )
   elsif operator = GREATER_THAN then
      return( { EVAL_SUCCESS, o1 > o2 } )
   elsif operator = EQUAL_TO then
      return( { EVAL_SUCCESS, o1 = o2 } )
   elsif operator = CONCATENATE then
      return( { EVAL_SUCCESS, o1 & o2 } )
   elsif operator = AND then
      return( { EVAL_SUCCESS, o1 and o2 } )
   elsif operator = OR then
      return( { EVAL_SUCCESS, o1 or o2 } )
   elsif operator = LESS_OR_EQUAL then
      return( { EVAL_SUCCESS, o1 <= o2 } )
   elsif operator = GREATER_OR_EQUAL then
      return( { EVAL_SUCCESS, o1 >= o2 } )
   elsif operator = NOT_EQUAL then
      return( { EVAL_SUCCESS, o1 != o2 } )
   elsif operator = COMPARISON then
      return( { EVAL_SUCCESS, not compare( o1, o2 ) } )
   elsif operator = BITWISE_AND then
      return( { EVAL_SUCCESS, and_bits( o1, o2 ) } )
   elsif operator = BITWISE_OR then
      return( { EVAL_SUCCESS, or_bits( o1, o2 ) } )
   elsif operator = BITWISE_XOR then
      return( { EVAL_SUCCESS, xor_bits( o1, o2 ) } )
   elsif operator = POWER then
      return( { EVAL_SUCCESS, power( o1, o2 ) } )
   elsif operator = MOD then
      if recfind( 0, o2 ) then
	 return( { EVAL_ERROR, "ERROR: Attempt to divide by zero." } )
      end if
      return( { EVAL_SUCCESS, remainder( o1, o2 ) } )
   end if
end function


function resolve_unary_ops( sequence s, sequence s2 )
   --handles all unary operations
   integer op, loc    --operation, location

   if s2[1] = 'o' then
      --first item is operator, must be a unary op
      loc = 1
   else
      --two operators in a row, second one must be a unary op
      loc = match( "oo", s2 )
   end if

   while loc do
      if loc = 1 then
	 --correct for fact that it's not second in a pair of operators
	 loc = 0
      end if
      op = find( s[loc + 1], UNARY_OPS )
      if not op then
	 --ERROR: Expected unary operator, not s[loc + 1].
	 return( {EVAL_ERROR, "Error: expected unary operator, not '" & s[loc+1] & "'." } )
      end if
      if op = PLUS then
	 --unary + does nothing...
      elsif op = MINUS then
	 --arithmetical negation
	 s[loc + 2] = - s[loc + 2]
      elsif op = NOT then
	 --logical negation
	 s[loc + 2] = not s[loc + 2]
      elsif op = BITWISE_NOT then
	 --bitflip
	 s[loc + 2] = not_bits( s[loc + 2] )
      end if
      --remove operator
      s = s[1..loc] & s[loc + 2..length( s )]
      s2 = s2[1..loc] & s2[loc + 2..length( s2 )]

      --see if there's another unary operator in the expression
      loc = match( "oo", s2 )
   end while

   return( {EVAL_SUCCESS, { s, s2 } } )
end function


function resolve_binary_ops( sequence s, sequence s2, sequence ops )
   --handles all binary operations within the given set
   --(ops, which is a set of constants)
   --returns {EVAL_SUCCESS, {s, s2} } or {EVAL_ERROR, errormsg}
   object ufo       --operation result
   integer ci, o    --count integer, operator index
   ci = 1
   while ci < length( s2 ) do
      if s2[ci] = 'o' then --operator
	 o = find( s[ci], OPERATORS )
	 if find( o, ops ) then
	    --it's one we're looking for
	    ufo = operate( s[ci - 1], o, s[ci + 1] )
	    if ufo[1] = EVAL_SUCCESS then
	       --replace both operands and the operator with the result
	       s = s[1..ci - 2] & { ufo[2] } & s[ci + 2..length( s )]
	       s2 = s2[1..ci - 2] & 't' & s2[ci + 2..length( s2 )]
	       ci = ci - 1
	    else
	       return( ufo )
	    end if
	 end if
      end if
      ci = ci + 1
   end while
   return( { EVAL_SUCCESS, { s, s2 } } )
end function


function form_sequences( sequence s, sequence s2 )
   --forms sequences
   --    have:
   --    s =  { "{",  4,  ",",  5,  ",",  6,  "}" }
   --    s2 = { '{', 't', ',', 't', ',', 't', '}' } = "{t,t,t}"
   --    want:
   --    s =  { {4,5,6} }
   --    s2 = {   't'   } = "t"
   object res  --recursion result
   integer ci, b, e, inseq, b2, e2
   sequence val

   ci = 1     --count integer
   b = 0      --beginning index of sequence
   e = 0      --ending index of sequence
   inseq = 0  --inside sequence?
   val = {}   --spliced sequence

   while ci <= length( s2 ) do
      if s2[ci] = '{' then
	 if inseq then
	    --recurse and replace
	    b2 = ci
	    e2 = find( '}', s2[b2 + 1..length( s2 )] ) + b2
	    res = form_sequences( s[b2..e2], s2[b2..e2] )
	    s = s[1..b2 - 1] & res[1] & s[e2 + 1..length( s )]
	    s2 = s2[1..b2 - 1] & res[2] & s2[e2 + 1..length( s2 )]
	    ci = b2
	 else
	    b = ci
	    inseq = 1
	    val = {}
	 end if
      elsif s2[ci] = '}' then
	 e = ci
	 inseq = 0
	 s = s[1..b - 1] & { val } & s[e + 1..length( s )]
	 s2 = s2[1..b - 1] & 't' & s2[e + 1..length( s2 )]
	 ci = b
      elsif s2[ci] = 't' then
	 if inseq then
	    val = append( val, s[ci] )
	 end if
      end if
      ci = ci + 1
   end while

   return( { s, s2 } )
end function


-----------------------------------------------------------------------------
-- Major functions
-----------------------------------------------------------------------------

function parse_expression( sequence s )
   --breaks up a string into 'words' (a setofstrings)
   --keeps string literals, character literals, and dual-operators together
   sequence sb, word
   integer lo, quoted, escd
   sb = {}
   word = {}
   lo = 0
   quoted = 0
   escd = 0

   for cc = 1 to length( s ) do
      if quoted then   --inside a string or character literal
	 if s[cc] = quoted then  --character is right type of quote
	    if escd then    --last character was a '\\', so quote is escaped
	       word = word & s[cc]
	       escd = 0
	    else
	       --closing quote
	       word = word & s[cc]
	       sb = sb & { word }
	       word = {}
	       quoted = 0
	    end if
	 else
	    word = word & s[cc]
	    if s[cc] = '\\' then
	       escd = not escd
	    else
	       escd = 0
	    end if
	 end if
      else --not in a string
	 if separator( s[cc] ) then
	    if find( lo & s[cc], OPERATORS ) then --dual character operator
	       sb[length( sb )] = sb[length( sb )] & s[cc]
	       word = {}
	       lo = 0
	    else
	       if length( word ) then
		  sb = sb & { word }
	       end if
	       word = {}
	       if not find( s[cc], WHITESPACE ) then
		  --ignore whitespace since we're not inside a string
		  if s[cc] = '"' then
		     --opening string quote
		     word = { s[cc] }
		     lo = 0
		     quoted = '"'
		  elsif s[cc] = '\'' then
		     --opening character quote
		     word = { s[cc] }
		     lo = 0
		     quoted = '\''
		  else
		     --single-character operator or symbol
		     sb = sb & { { s[cc] } }
		     lo = s[cc]
		  end if
	       end if
	    end if
	 else
	    --part of a literal or identifier
	    word = word & s[cc]
	    lo = 0
	 end if
      end if
   end for
   if length( word ) then
      sb = sb & { word }
   end if
   if quoted then
      return( {EVAL_ERROR, "Error: Invalid string or character literal, mismatched quotes."} )
   end if
   return( {EVAL_SUCCESS, sb} )
end function


function preeval( sequence s )
   --s is a parsed expression
   --checks format of expression and validity of terms and operators
   --replaces terms with their values
   --builds semi-tokenized parallel sequence
   object l         --literal value result
   integer tc, snl  --term count, sequence nesting level
   sequence s2      --semitokenized sequence
   tc = 0
   snl = 0
   s2 = repeat( 0, length( s ) )

   for c = 1 to length( s ) do
      if find( s[c], OPERATORS ) then
	 --operators OK except at end of expression
	 s2[c] = 'o'
	 if c = length( s ) then
	    return( { EVAL_ERROR, "ERROR: Expression ends with operator, expected another term." } )
	 end if
      elsif not compare( s[c], "(" ) then
	 --subexpression parentheses OK
	 s2[c] = '('
      elsif not compare( s[c], ")" ) then
	 --subexpression parentheses OK
	 s2[c] = ')'
      elsif not compare( s[c], "{" ) then
	 --increase sequence nest level
	 snl = snl + 1
	 s2[c] = '{'
      elsif not compare( s[c], "," ) then
	 --commas OK only if inside sequence brackets
	 if not snl then
	    return( { EVAL_ERROR, "ERROR: Commas only allowed in sequence literals." } )
	 end if
	 s2[c] = ','
      elsif not compare( s[c], "}" ) then
	 --decrease sequence nest level
	 snl = snl - 1
	 s2[c] = '}'
      else
	 l = numeric_literal( s[c] )
	 if l[1] then --number
	    s[c] = l[2] --replace the term with it's value
	    s2[c] = 't'
	 else
	    l = string_or_character_literal( s[c], '"' )
	    if l[1] then --string
	       s[c] = l[2] --replace the term with it's value
	       s2[c] = 't'
	    else
	       l = string_or_character_literal( s[c], '\'' )
	       if l[1] then --character
		  s[c] = l[2] --replace the term with it's value
		  s2[c] = 't'
	       else
		  --not an operator or literal!
		  return( { EVAL_ERROR, "ERROR: Invalid term or operator in expression." } )
	       end if
	    end if
	 end if
      end if
   end for

   --make sure there are terms
   if not find( 't', s2 ) then
      return( { EVAL_ERROR, "ERROR: Expression contains no terms." } )
   end if

   --check for sneaky sequence formation errors
   if snl != 0 then
      return( { EVAL_ERROR, "ERROR: Invalid sequence literal formation, mismatched {}s." } )
   end if
   if match( ",}", s2 ) then
      return( { EVAL_ERROR, "ERROR: Improper sequence formation, comma after last element." } )
   elsif match( "o,", s2 ) or match( "o}", s2 ) then
      return( { EVAL_ERROR, "ERROR: Sequence element ends with operator." } )
   end if

   return( { EVAL_SUCCESS, {s, s2} } )
end function


function evaluate( sequence inp )
   --s is an expression in parsed, preeval'ed form
   --returns {EVAL_SUCCESS, value} or {EVAL_ERROR, errormsg}
   sequence s, s2  --sequence, semi-tokenized sequence
   object ufo
   integer ci   --count integer

   s = inp[1]
   s2 = inp[2]
   ufo = 0

   --break out now if we've got the answer
   if not compare( s2, "t" ) then
      return( { EVAL_SUCCESS, s } )
   end if

   --parenthesize sequence elements
   if find( '{', s2 ) then
      --skip if no sequences
      ci = 1
      while ci <= length( s2 ) do
	 if s2[ci] = '{' then
	    --insert opening parenthese at start of element
	    s = s[1..ci] & {"("} & s[ci + 1..length( s )]
	    s2 = s2[1..ci] & '(' & s2[ci + 1..length( s2 )]
	 elsif s2[ci] = ',' then
	    --insert closing parenthese at end of element and
	    --opening parenthese at start of next element
	    s = s[1..ci - 1] & {")", ",", "("} & s[ci + 1..length( s )]
	    s2 = s2[1..ci - 1] & "),(" & s2[ci + 1..length( s2 )]
	    ci = ci + 1
	 elsif s2[ci] = '}' then
	    --insert closing parenthese at end of element
	    s = s[1..ci - 1] & {")"} & s[ci..length( s )]
	    s2 = s2[1..ci - 1] & ')' & s2[ci..length( s2 )]
	    ci = ci + 1
	 end if
	 ci = ci + 1
      end while
   end if
   --

   --scan for subexpressions
   if find( '(', s2 ) or find( ')', s2 ) then
      --skip if no subexpressions
      ci = 1
      ufo = {0, 0, 0, 0}  --{nestinglevel, start, end, result}
      while ci <= length( s2 ) do
	 if s2[ci] = '(' then
	    if not ufo[1] then
	       ufo[2] = ci --subexpression start
	    end if
	    ufo[1] = ufo[1] + 1 --inc nesting level
	 elsif s2[ci] = ')' then
	    ufo[1] = ufo[1] - 1 --dec nesting level
	    if not ufo[1] then
	       ufo[3] = ci --subexpression end
	       ufo[4] = evaluate( { s[ufo[2] + 1..ufo[3] - 1],
				   s2[ufo[2] + 1..ufo[3] - 1] } )
	       if ufo[4][1] != EVAL_SUCCESS then
		  --error in subexpression
		  return( ufo[4] )
	       end if
	       s = s[1..ufo[2] - 1] &
		   ufo[4][2] &
		   s[ufo[3] + 1..length( s2 )]
	       s2 = s2[1..ufo[2] - 1] &
		   't' &
		   s2[ufo[3] + 1..length( s2 )]
	       ci = ufo[2]
	    end if
	 end if
	 ci = ci + 1
      end while
      if ufo[1] then
	 --ERROR: Mismatched parentheses.
	 return( { EVAL_ERROR, "ERROR: Mismatched parentheses." } )
      end if
      --break out now if we've got the answer
      if not compare( s2, "t" ) then
	 return( { EVAL_SUCCESS, s } )
      end if
   end if
   --

   --sequence elements should have been evaluated by now, so we can form
   --sequences up into terms
   if find( '{', s2 ) then
      --skip if no sequences
      ufo = form_sequences( s, s2 )
      s = ufo[1]
      s2 = ufo[2]
      --break out now if we've got the answer
      if not compare( s2, "t" ) then
	 return( { EVAL_SUCCESS, s } )
      end if
   end if
   --

   --scan for unary operators
   ufo = resolve_unary_ops( s, s2 )
   if ufo[1] != EVAL_SUCCESS then
      return( ufo ) --error
   end if
   s = ufo[2][1]
   s2 = ufo[2][2]
   --break out now if we've got the answer
   if not compare( s2, "t" ) then
      return( { EVAL_SUCCESS, s } )
   end if
   --

   --scan for binary ops
   for level = 1 to length( PRECEDENCE ) do
      ufo = resolve_binary_ops( s, s2, PRECEDENCE[level] )
      if ufo[1] != EVAL_SUCCESS then
	 return( ufo ) --error
      end if
      s = ufo[2][1]
      s2 = ufo[2][2]
      --break out now if we've got the answer
      if not compare( s2, "t" ) then
	 return( { EVAL_SUCCESS, s } )
      end if
   end for
   --

   --should never get here...but just in case:
   return( { EVAL_ERROR, "ERROR: Expression contains unknown operators or invalid terms." } )
end function


global function eval( sequence s )
   -- s is an expression in string form
   --parses, preevaluates, and evaluates the expression
   object o

   if length( s ) then
      o = parse_expression( s )
      if o[1] then
	 o = preeval( o[2] )
	 if o[1] then
	    return( evaluate( o[2] ) )
	 end if
      end if
      return( o )
   end if
   return( { EVAL_ERROR, "ERROR: Null expression string, nothing to evaluate." } )
end function


-----------------------------------------------------------------------------
-- End eval.e
----------------------------------------------------------------------------



